#!/usr/local/bin/python3.5
# Copyright (c) 2017 David Preece, All rights reserved.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import json
from base64 import b64encode
from tfnz.cli import generic_cli, base_argparse


def main():
    parser = base_argparse('tfresources')
    generic_cli(parser, {None: list_resources})


def list_resources(location, args):
    resources = {
        'location': location.location,
        'nodes': {b64encode(node.pk).decode(): node.stats for node in location.nodes.values()},
        'volumes': [vol.namespaced_display_name() for vol in location.volumes.values()],
        'endpoints': [ep.domain for ep in location.endpoints.values()]
    }

    print(json.dumps(resources, indent=2))


if __name__ == "__main__":
    main()



# Copyright (c) 2017 David Preece, All rights reserved.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

# A note on nomenclature
# A tunnel logically goes from localhost:port to container:port but has no TCP connections
# The tunnel is made from zero or more proxies which do the actual tunnelling

import logging
import socket
import time
import shortuuid
import weakref
from . import Killable


class Tunnel(Killable):
    """An object representing a TCP proxy from localhost onto a container.
    Do not instantiate directly, use location.tunnel_onto or location.wait_http_200.

    Interact with the proxy through TCP (or call localport if you didn't set it explicitly).
    Note that apparently plaintext traffic through the tunnel is still encrypted on the wire."""

    def __init__(self, connection, node, container, port, lp=None, bind=None, timeout=30):
        super().__init__()
        # tell the location what we want
        self.uuid = shortuuid.uuid().encode()
        self.sess = connection.rid
        self.node = node
        self.connection = weakref.ref(connection)
        self.socket = None
        self.container = weakref.ref(container)
        self.port = port
        self.lp = lp  # can be None
        self.bind = bind  # can be None
        self.fd = None
        self.tcpip_direct_return = None
        self.timeout = timeout
        self.proxies = {}
        self.created = False
        connection.register_connect_callback(self._session_reconnected)

    def connect(self):
        self.ensure_alive()
        # we need to be able to call connect twice...
        if self.created:
            return
        self.created = True
        # create the listen socket
        self.socket = socket.socket()
        self.socket.setblocking(False)
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.socket.bind(('0.0.0.0' if self.bind is None else self.bind, 0 if self.lp is None else self.lp))
        self.lp = self.socket.getsockname()[1]
        self.socket.listen()
        self.fd = self.socket.fileno()
        self.connection().loop.register_exclusive(self.fd, self.event, comment="listener for " + str(self))

        # have the location create its end.
        self.connection().send_cmd(b'create_tunnel',
                                   {'container': self.container().uuid,
                                    'port': self.port,
                                    'timeout': self.timeout}, uuid=self.uuid)
        logging.info("Creating remote tunnel: %s (%d -> %d)" % (self.uuid.decode(), self.lp, self.port))

    def connect_tcpip_direct(self, caller):
        self.ensure_alive()
        # we need to be able to call connect twice...
        if self.created:
            return
        self.created = True
        self.tcpip_direct_return = caller

        # have the location create its end.
        self.connection().send_cmd(b'create_tunnel',
                                   {'container': self.container().uuid,
                                    'port': self.port,
                                    'timeout': self.timeout}, uuid=self.uuid)
        logging.info("Creating remote tcpip direct tunnel: %s (port %d)" % (self.uuid.decode(), self.port))

    def stdin(self, data):
        # Direct tcpip
        # Is this a close?
        if len(data) == 0:
            self.destroy()
            self.tcpip_direct_return.tunnel_closed(self)
        else:
            self.connection().send_cmd(b'to_proxy', {"tunnel": self.uuid,
                                                     "proxy": 0}, bulk=data)

    def localport(self) -> int:
        """Returns the (possibly dynamically allocated) local port number.

        :return: the port number as an int"""
        self.ensure_alive()
        return self.lp

    def destroy(self, with_command=True):
        # Destroy this tunnel.
        if self.bail_if_dead():
            return
        self.mark_as_dead()
        self.connection().unregister_connect_callback(self._session_reconnected)
        self.disconnect_all_proxies()
        if self.fd is not None:
            self.connection().loop.unregister_exclusive(self.fd)
        if self.socket is not None:
            self.socket.close()
        if with_command:
            self.connection().send_cmd(b'destroy_tunnel', {"tunnel": self.uuid})
        logging.info("Destroyed remote tunnel: " + self.uuid.decode())

    def disconnect_all_proxies(self):
        # when the container reboots we need to close the tcp connections
        for proxy in list(self.proxies.items()):
            self.connection().loop.unregister_exclusive(proxy[0])
            proxy[1].close()
        self.proxies.clear()

    def event(self, localfd):
        # An event on the connection itself
        # The accept call sometimes fails
        if self.bail_if_dead():
            return
        new_proxy = None
        while new_proxy is None:
            try:
                new_proxy = self.socket.accept()
            except BlockingIOError:
                logging.debug("Attempting to accept a new connection but the resource was temporarily unavailable")
                time.sleep(0.1)

        fd = new_proxy[0].fileno()
        self.proxies[fd] = new_proxy[0]
        self.connection().loop.register_exclusive(fd, self.to_proxy, comment="proxy fd=" + str(fd))

        # send nothing to force the connection open on the far end - some servers like to talk first
        self.connection().send_cmd(b'to_proxy', {"tunnel": self.uuid,
                                                 "proxy": fd}, bulk=b'')

        logging.debug("Accepted proxy connection, fd: " + str(fd))

    def to_proxy(self, localfd):
        # Send to the location which will forward to the end client.
        if self.bail_if_dead():
            return
        try:
            data = b''
            try:
                data = self.proxies[localfd].recv(8192)
            except ConnectionResetError:
                logging.debug("Connection reset when receiving from proxy")
                pass
            if data == b'':
                logging.debug("Received no data, assuming socket was closed for proxy: " + str(localfd))
                self.close_proxy(localfd)
                self.connection().send_cmd(b'close_proxy', {"tunnel": self.uuid,
                                                            "proxy": localfd})
            else:
                logging.debug("Sending data to proxy: " + str(localfd))
                self.connection().send_cmd(b'to_proxy', {"tunnel": self.uuid,
                                                         "proxy": localfd}, bulk=data)
        except KeyError:
            logging.debug("Received a forwarding request to a proxy not in map: " + str(localfd))

        return True  # everything went fine

    def from_proxy(self, msg):
        # Data being sent from the container
        if self.bail_if_dead():
            return

        # is this a tcpip_direct tunnel?
        if self.tcpip_direct_return is not None:
            self.tcpip_direct_return.data(self, msg.bulk)
            return

        # headed for a tcp proxy as usual
        try:
            proxy_fd = msg.params['proxy']
        except KeyError:
            # a blank message with no proxy id is to let us know it constructed server side
            logging.debug("From proxy message with no proxy, marking tunnel as ready: " + self.uuid.decode())
            return

        if msg.command == 'close_proxy':
            logging.debug("Server told us to close connection: " + str(proxy_fd))
            self.close_proxy(proxy_fd)
            return

        try:
            skt = self.proxies[proxy_fd]
            skt.sendall(msg.bulk)
        except KeyError:
            pass  # proxy has already gone away

    def close_proxy(self, msg_or_fd):
        if self.bail_if_dead():
            return

        # destroy then inform the ssh server
        if self.tcpip_direct_return is not None:
            self.destroy()
            self.tcpip_direct_return.close(self)
            return

        # Close one single proxy
        try:
            try:
                fd = msg_or_fd.params['proxy']
            except AttributeError:
                fd = msg_or_fd  # when we close the proxy locally we will have been passed a file desriptor
            self.proxies[fd].close()
            self.connection().loop.unregister_exclusive(fd)
            del self.proxies[fd]
            logging.debug("Closed proxy connection, fd: " + str(fd))
        except KeyError:
            pass  # proxy has already gone away

    def _session_reconnected(self, rid):
        logging.debug("Tunnel reset its session id: " + self.uuid.decode())
        self.sess = rid

    def __repr__(self):
        return "<Tunnel '%s' localport=%d container=%s)>" % \
               (self.uuid.decode(), self.port, self.container().uuid.decode())



# Copyright (c) 2017 David Preece, All rights reserved.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import sys
import requests.exceptions
import json
import logging
from typing import Optional


class Docker:
    docker_socket = '/var/run/docker.sock'
    docker_url_base = 'http+unix://%2Fvar%2Frun%2Fdocker.sock'
    session = None

    @staticmethod
    def description(docker_image_id: str, *, conn: Optional['Connection']=None) -> dict:
        """Describe a docker image.

        :param docker_image_id: Docker image id.
        :param conn: An optional connection to the location.
        :return: A dict representation of image metadata."""
        # try locally
        can_connect_local = True
        try:
            r = Docker._session().get('%s/images/%s/json' % (Docker.docker_url_base, docker_image_id))

            # local docker works but doesn't have it
            if r.status_code == 404:
                logging.info("Local docker doesn't have image, trying for remote")
            else:
                # presumably worked, cache it and return
                descr = json.loads(r.text)

                # strip some stuff we don't need
                removes = ('Container', 'Comment', 'ContainerConfig', 'GraphDriver')
                for remove in removes:
                    if remove in descr:
                        del descr[remove]

                # all good
                return descr
        except requests.exceptions.ConnectionError:
            can_connect_local = False

        # no go locally, try remotely
        if conn is not None:
            logging.info("Retrieving description: " + docker_image_id)
            msg = conn.send_blocking_cmd(b'retrieve_description', {'image_id': docker_image_id})
            if 'description' in msg.params:
                return msg.params['description']

        # image is in neither location
        if can_connect_local:
            if conn is not None:
                raise RuntimeError("Cannot find image in either local docker or remote image cache: " + docker_image_id)
            else:
                raise RuntimeError("Cannot find image in local docker: " + docker_image_id)
        else:
            # local's dead, too
            Docker._docker_warning()

    @staticmethod
    def tarball(docker_image_id: str) -> bytes:
        """Retrieve the tarball of a docker image.

        :param docker_image_id: Docker image id.
        :return: A stream of bytes that would be the contents of the tar archive."""
        try:
            r = Docker._session().get('%s/images/%s/get' % (Docker.docker_url_base, docker_image_id))
            return r.content
        except requests.exceptions.ConnectionError:
            Docker._docker_warning()


    @staticmethod
    def last_image() -> str:
        """Finding the most recent docker image on this machine.

        :return: Docker image id of the most recently built docker image"""
        r = None
        try:
            r = Docker._session().get('%s/images/json' % Docker.docker_url_base)
        except requests.exceptions.ConnectionError:
            Docker._docker_warning()
        if len(r.text) == 0:
            raise ValueError("Docker has no local images.")
        obj = json.loads(r.text)
        return obj[0]['Id'][7:19]

    @staticmethod
    def _docker_warning():
        print("""
    Cannot (and need to) connect to the docker socket
    -------------------------------------------------
    
    The remote cache does not have a description for the combination of this image and this user.
    If you think it should, have changed you which user account you're using?
    Is docker running on this machine? 
    You may need to run sudo chmod 666 /var/run/docker.sock
    """, file=sys.stderr)
        raise RuntimeError("Need a functioning local Docker")

    @staticmethod
    def _session():
        # when we need unix sockets (not deployed on server, hence late binding)
        if Docker.session is None:
            import requests_unixsocket
            Docker.session = requests_unixsocket.Session()
        return Docker.session



# Copyright (c) 2017 David Preece, All rights reserved.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import weakref
import logging
import requests
import time
from requests.exceptions import ConnectionError, ReadTimeout
from typing import List, Optional, Tuple
from tfnz.container import Container


class Cluster:
    """An object representing a collection of containers, load balanced and published to an endpoint"""

    def __init__(self, *, containers: Optional[List[Container]]=None, rewrite: Optional[str]=None):
        """Create a cluster.

        :param containers: An optional list of containers to initialise the cluster with.
        :param rewrite: An optional string to be rewritten into the http host header."""
        self.uuid = None
        self.conn = None
        self.containers = {}
        self.rewrite = rewrite
        for container in containers:
            self.add_container(container)

    def add_container(self, container):
        """Add a container to the cluster.

        :param container: the container to add."""
        if container.uuid in self.containers:
            pass
        container.wait_until_ready()
        self.containers[container.uuid] = container

        # let the server know, maybe
        if self.conn is not None:
            self.conn().send_blocking_cmd(b'add_to_cluster', {'cluster': self.uuid,
                                                              'container': container.uuid})

    def remove_container(self, container):
        """Remove a container from the cluster.

        :param container: the container to remove."""
        try:
            del self.containers[container.uuid]

            # let the server know, maybe
            if self.conn is not None:
                self.conn().send_cmd(b'remove_from_cluster', {'cluster': self.uuid,
                                                              'container': container.uuid})
        except KeyError:  # was not in the list of containers
            pass

    def uuids(self):
        return self.containers.keys()

    def __repr__(self):
        return "<Cluster '%s' containers=%d>" % (self.uuid, len(self.containers))


class WebEndpoint:
    """An HTTP proxy that can expose a number of clusters"""

    def __init__(self, conn: 'Connection', domain: str):
        """Do not construct directly, see location.endpoints"""
        self.conn = weakref.ref(conn)
        self.domain = domain
        self.clusters = {}  # uuid to cluster

    def publish(self, cluster: Cluster, fqdn: str, *, ssl: Optional[Tuple]=None):
        """Publish a cluster onto an http/https endpoint.
        To update a cluster, merely re-publish onto the same endpoint.

        :param cluster: The cluster to publish.
        :param fqdn: The fqdn to publish.
        :param ssl: A tuple of (cert.pem, key.pem) or (cert.pem, key.pem, cert.intermediate)."""
        # checks
        if not fqdn.endswith(self.domain):
            raise ValueError("Web endpoint for (%s) cannot publish: %s" % (self.domain, fqdn))
        if cluster.uuid in self.clusters:  # already published
            return

        # ssl check and read
        combined = None
        if ssl is not None:
            if len(ssl) not in (2, 3):
                raise ValueError("SSL needs to be a tuple of (cert.pem, key.pem) or "
                                 "(cert.pem, key.pem, cert.intermediate")
            combined = ''
            with open(ssl[0]) as f:
                combined += f.read()
            with open(ssl[1]) as f:
                combined += f.read()
            if len(ssl) is 3:
                with open(ssl[2]) as f:
                    combined += f.read()

        # tell the location to publish
        subdomain = fqdn[:-len(self.domain)]
        msg = self.conn().send_blocking_cmd(b'publish_web', {'domain': self.domain,
                                                             'subdomain': subdomain,
                                                             'rewrite': cluster.rewrite,
                                                             'ssl': combined,
                                                             'containers': list(cluster.uuids())})
        logging.info("Published (%s) at: %s" % (msg.uuid.decode(), subdomain + self.domain))

        cluster.uuid = msg.uuid
        cluster.conn = weakref.ref(self.conn())
        self.clusters[msg.uuid] = cluster
        return msg.uuid

    @staticmethod
    def wait_http_200(fqdn: str, *, ssl: Optional[bool]=False):
        """Poll the gateway for an http 200 from this cluster.

        :param fqdn: the fqdn to poll.
        :param ssl: optionally connect via ssl."""
        url = '%s://%s' % ('https' if ssl else 'http', fqdn)
        attempts_remaining = 30
        while True:
            try:
                r = requests.get(url, timeout=5)
                if r.status_code == 200:
                    break
            except (ConnectionError, ConnectionRefusedError, ReadTimeout):
                pass
            attempts_remaining -= 1
            if attempts_remaining == 0:
                raise ValueError("Could not connect to: " + url)
            time.sleep(1)

    def unpublish(self, cluster: Cluster):
        """Remove a cluster from a web endpoint.

        :param cluster: the cluster to remove."""
        if cluster.uuid is None or cluster.uuid not in self.clusters:  # not published anyway
            return

        self.conn().send_cmd(b'unpublish_web', {'cluster': cluster.uuid})
        logging.info("Unpublished: " + cluster.uuid.decode())
        cluster.conn = None
        del self.clusters[cluster.uuid]

    def __repr__(self):
        return "<WebEndpoint '%s' clusters=%d>" % (self.domain, len(self.clusters))



# Copyright (c) 2017 David Preece, All rights reserved.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import logging
import shortuuid
import weakref
from typing import Optional, List, Union, Callable, Any, Tuple
from . import Waitable, Killable, Connectable
from .tunnel import Tunnel
from .process import Process
from .ssh import SshServer


class Container(Waitable, Killable, Connectable):
    """An object representing a single container. Do not instantiate directly, use node.spawn."""
    def __init__(self, parent, image, uuid, docker_config, env, volumes, *, stdout_callback, termination_callback):
        Waitable.__init__(self)
        Killable.__init__(self)
        Connectable.__init__(self, parent.conn(), uuid, parent, None)
        self.parent = weakref.ref(parent)
        self.location = weakref.ref(self.parent().parent())
        self.image = image
        self.processes = {}
        self.ssh_servers = {}
        self.docker_config = docker_config
        self.env = env
        self.volumes = volumes
        self.stdout_callback = stdout_callback
        self.termination_callback = termination_callback

    def start(self):
        """Start a container that was spawned with sleep=True"""
        self.conn().send_cmd(b'wake_container', {'node': self.parent().pk, 'container': self.uuid})

    def private_ip(self):
        """Reports the container's ip address"""
        self.ensure_alive()
        self.wait_until_ready()
        return self.ip

    def node(self):
        """Returns the node object this container is running on"""
        return self.parent()

    def stdin(self, data: bytes):
        """Writes the data into the container's stdin.

        :param data: The data to be written."""
        self.ensure_alive()
        self.wait_until_ready()
        self.conn().send_cmd(b'stdin_container', {'node': self.parent().pk,
                                                  'container': self.uuid},
                             bulk=data)

    def wait_tcp(self, dest_port):
        """Connects and disconnects a single connection onto this container and a given port.

        If your sevice is only capable of .accept'ing once, do not use this.

        :param dest_port: destination tcp port < 1024 is fine."""
        self.ensure_alive()
        self.wait_until_ready()
        self.location()._wait_tcp(self, dest_port)

    def attach_tunnel(self, dest_port: int, *, localport: Optional[int]=None, bind: Optional[str]=None) -> Tunnel:
        """Creates a TCP proxy between localhost and a container.

        :param dest_port: The TCP port on the container to connect to.
        :param localport: Optional choice of local port no - if not provided uses the dest_port.
        :param bind: Optionally bind to an address other than localhost.
        :return: A Tunnel object.

        This call does no checking to ensure the server side is ready -
        but a failed connection will not destroy the tunnel itself and will poll until connected.
        """
        self.ensure_alive()
        localport = dest_port if localport is None else localport
        return self.location()._tunnel_onto(self, dest_port, localport, bind)

    def wait_http_200(self, *, dest_port: Optional[int]=80, fqdn: Optional[str]='localhost', path: Optional[str]='') \
            -> Tunnel:
        """Poll until an http 200 is returned.

        :param dest_port: Override the default port.
        :param fqdn: A host name to use in the http request.
        :param path: A path on the server - appended to /
        :return: A Tunnel object.
        """
        self.ensure_alive()
        self.wait_until_ready()
        return self.location()._wait_http_200(self, dest_port, fqdn, path)

    def destroy_tunnel(self, tunnel: Tunnel):
        """Destroy a tunnel

        :param tunnel: The tunnel to be destroyed."""
        self.ensure_alive()
        self.location()._destroy_tunnel(tunnel, container=self)

    def all_tunnels(self) -> List[Tunnel]:
        """Returns all the tunnels connected to this container

        :return: A list of Tunnel objects"""
        self.ensure_alive()
        return [t for t in self.location().tunnels.values() if t.container() == self]

    def allow_connection_from(self, container: 'Container'):
        """Allow another container to call this one over private ip

        :param container: The container that will be allowed to call."""
        self.ensure_alive()
        self.wait_until_ready()
        container.wait_until_ready()
        super().allow_connection_from(container)

    def disallow_connection_from(self, container: 'Container'):
        """Stop allowing another container to call this one over private ip

        :param container: The container that will no longer be allowed to call."""
        if self.bail_if_dead():
            return
        super().disallow_connection_from(container)

    def spawn_process(self, remote_command: str,
                      data_callback: Optional[Callable]=None,
                      termination_callback: Optional[Callable]=None,
                      stderr_callback: Optional[Callable]=None) -> Process:
        """Spawn a process within a container, receives data asynchronously via a callback.

        :param remote_command: The command to remotely launch as a string (i.e. not list).
        :param data_callback: A callback for arriving data - signature (object, bytes).
        :param termination_callback: For when the process completes - signature (object), returncode.
        :param stderr_callback: For data emitted by the process's stderr stream - signature (object, bytes).
        :return: A Process object.
        """
        if isinstance(remote_command, list):
            raise ValueError("Pass the command as a single string (that gets passed to a shell), not a list")
        self.ensure_alive()
        self.wait_until_ready()

        # Cool, go.
        logging.info("Container (%s) spawning process: '%s'" % (self.uuid.decode(), remote_command))

        # get the node to launch the process for us
        # we need the uuid of the spawn command because it's used to indicate when the process has terminated
        spawn_command_uuid = shortuuid.uuid().encode()
        self.processes[spawn_command_uuid] = Process(self, spawn_command_uuid,
                                                     data_callback,
                                                     termination_callback,
                                                     stderr_callback)
        self.conn().send_cmd(b'spawn_process', {'node': self.parent().pk,
                                                'container': self.uuid,
                                                'command': remote_command},
                             uuid=spawn_command_uuid,
                             reply_callback=self._process_callback)
        logging.info("...process id: " + spawn_command_uuid.decode())
        return self.processes[spawn_command_uuid]

    def run_process(self, remote_command: str, *,
                    data_callback: Optional[Callable]=None,
                    stderr_callback: Optional[Callable]=None,
                    nolog: Optional[bool]=False) -> (bytes, bytes, str):
        """Run a process once, synchronously, without pty.

        :param remote_command: The command to run remotely.
        :param data_callback: A callback for arriving data - signature (object, bytes).
        :param stderr_callback: For data emitted by the process's stderr stream - signature (object, bytes).
        :param nolog: Don't log this command (to hide sensitive data).
        :return: stdout from the process, stderr from the process, exit code (as string).
        """
        if isinstance(remote_command, list):
            raise ValueError("Pass single shot commands a string.")
        self.ensure_alive()
        self.wait_until_ready()

        # Cool, go.
        if not nolog:
            logging.info("Container (%s) running process: '%s'" % (self.uuid.decode(), remote_command))

        # get the node to launch the process for us
        # we need the uuid of the spawn command because it's used to indicate when the process has terminated
        msg = self.conn().send_blocking_cmd(b'run_process', {'node': self.parent().pk,
                                                             'container': self.uuid,
                                                             'command': remote_command})
        if msg is None:
            raise ValueError("Blocking process failed: " + remote_command)
        if data_callback is not None and len(msg.params['stdout']) > 0:
            data_callback(self, msg.params['stdout'])
        if stderr_callback is not None and len(msg.params['stderr']) > 0:
            stderr_callback(self, msg.params['stderr'])
        return msg.params['stdout'], msg.params['stderr'], msg.params['exit_code']

    def spawn_shell(self, *,
                    data_callback: Optional[Callable]=None,
                    termination_callback: Optional[Callable]=None,
                    echo: bool=False) -> Process:
        """Spawn a shell within a container, expose as a process

        :param data_callback: A callback for arriving data - signature (object, bytes).
        :param termination_callback: For when the process completes - signature (object, returncode).
        :param echo: Whether or not the shell echoes input.
        :return: A Process object."""
        self.ensure_alive()
        self.wait_until_ready()

        # get the node to launch the process for us
        # we need the uuid of the spawn command because it's used to indicate when the process has terminated
        spawn_command_uuid = shortuuid.uuid().encode()
        self.processes[spawn_command_uuid] = Process(self, spawn_command_uuid, data_callback, termination_callback)
        self.conn().send_cmd(b'spawn_shell', {'node': self.parent().pk,
                                              'container': self.uuid,
                                              'echo': echo},
                             uuid=spawn_command_uuid,
                             reply_callback=self._process_callback)
        logging.info("Spawned shell: " + spawn_command_uuid.decode())
        return self.processes[spawn_command_uuid]

    def destroy_process(self, process: Process):
        """Destroy a process

        :param process: The process to be destroyed."""
        self.ensure_alive()
        if process.uuid not in self.processes:
            raise ValueError("Process is not apparently a child of this container")
        process.destroy()
        del self.processes[process.uuid]

    def all_processes(self) -> List[Process]:
        """Returns all the processes (that were manually launched) running on this container.

        :return: A list of Process objects"""
        self.ensure_alive()
        return list(self.processes.values())

    def create_ssh_server(self, port: int=2222) -> SshServer:
        """Create an ssh/sftp server on the given port.

        :param port: Local tcp port number
        :return: An SshServer object."""
        self.ensure_alive()
        self.wait_until_ready()
        s = SshServer(self, port)
        s.start()
        self.ssh_servers[s.uuid] = s
        return s

    def destroy_ssh_server(self, server):
        """Destroy an ssh/sftp server than belongs to this container.

        :param server: An SshServer object."""
        self.ensure_alive()
        if server.uuid not in self.ssh_servers:
            raise ValueError("Server does not belong to this container")
        server.stop()
        del self.ssh_servers[server.uuid]

    def fetch(self, filename: str) -> bytes:
        """Fetch a single file from the container.

        :param filename: The full-path name of the file to be retrieved.
        :return: the contents of the file as a bytes object.

        Since the file gets loaded into memory, this is a bad way to move large files (>1GB)."""
        self.ensure_alive()
        self.wait_until_ready()
        return self.conn().send_blocking_cmd(b'fetch_file', {'node': self.parent().pk,
                                                             'container': self.uuid,
                                                             'filename': filename}).bulk

    def put(self, filename: str, data: bytes):
        """Put a file into the container.

        :param filename: The full-path name of the file to be placed.
        :param data: The contents of the file as a bytes object.

        This will just overwrite so be careful. Note that new file paths are created on demand.
        Similarly to fetch, this is a bad way to move large files (>1GB).
        """
        self.ensure_alive()
        self.wait_until_ready()
        self.conn().send_blocking_cmd(b'put_file', {'node': self.parent().pk,
                                                    'container': self.uuid,
                                                    'filename': filename}, bulk=data)

    def reboot(self, *, reset_filesystem: Optional[bool]=False):
        """Synchronously reboot a container, optionally resetting the filesystem.

        :param reset_filesystem: Reset the container's filesystem to its 'as booted' state."""
        if self.dead:
            raise RuntimeError("Tried to restart a container but it has already been removed from the node")
        self.ensure_alive()
        self.wait_until_ready()  # in this case means it has been configured, prepared etc. once already

        # Restart
        logging.info("Restarting container: " + self.uuid.decode())
        self.mark_not_ready()  # so calls are forced to wait until the container reports it has rebooted
        self.conn().send_cmd(b'reboot_container', {'node': self.parent().pk,
                                                   'container': self.uuid,
                                                   'reset_filesystem': reset_filesystem},
                             reply_callback=self.parent()._container_status_update)
        self.wait_until_ready()

    def _internal_destroy(self, send_cmd=True):
        # Destroy this container
        if self.bail_if_dead():
            return
        self.ensure_alive()
        self.wait_until_ready()

        # Fake the processes being destroyed (they will be anyway)
        for proc in list(self.processes.values()):
            proc.destroy(with_command=False)

        # Destroy any tunnels
        for tun in list(self.all_tunnels()):
            self.location()._destroy_tunnel(tun, self, with_command=False)

        # Destroy any ssh servers
        for svr in list(self.ssh_servers.values()):
            self.destroy_ssh_server(svr)

        # Destroy (async)
        if send_cmd:
            logging.info("Destroying container: " + self.uuid.decode())
            self.conn().send_cmd(b'destroy_container', {'node': self.parent().pk,
                                                        'container': self.uuid})
        else:
            logging.info("Container has exited and/or been destroyed: " + self.uuid.decode())
        self.mark_as_dead()

        # Callback
        if self.termination_callback is not None:
            self.location().call_on_main(self.termination_callback, (self, 0))

    def _process_callback(self, msg):
        if self.bail_if_dead():
            return

        if msg.uuid not in self.processes:
            logging.debug("Message arrived for an unknown process: " + msg.uuid.decode())
            return

        logging.debug("Received data from process: " + msg.uuid.decode())
        logging.debug(msg.bulk)
        self.processes[msg.uuid].give_me_messages(msg)

    def __repr__(self):
        return "<Container `%s` image=%s ip=%s>" % (self.uuid.decode(), self.image, self.ip)


class ExternalContainer(Connectable):
    """An object representing a container managed by another session (and the same user) but advertised using a tag"""

    def __init__(self, conn, uuid, node, ip):
        super().__init__(conn, uuid, node, ip)

    def private_ip(self):
        """Reports the container's ip address"""
        return self.ip

    def __repr__(self):
        return "<ExternalContainer '%s' ip=%s>" % (self.uuid.decode(), self.ip)



# Copyright (c) 2018 David Preece, All rights reserved.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import random
import string
import logging
import time
from threading import Thread
from tfnz import Waitable
from tfnz.node import Node
from tfnz.volume import Volume
from tfnz.container import Container


class Postgresql(Waitable):
    """An object encapsulating a Postgresql server running on it's default port (5432).
        Connect with username=postgres.

        :param node: The node to spawn on.
        :param volume: A volume (object) to use as a persistent store.
        :param password: An optional password for the database, will create one if not supplied.
        :param log_callback: An optional callback for log messages -  signature (object, bytes)
        :param image: Specify a non-default image."""
    def __init__(self, node: Node, volume: Volume, *, password: str=None, log_callback=None, image=None):
        super().__init__()
        # passwords
        if password is None:
            self.password = ''.join(random.SystemRandom().choice(string.ascii_letters+string.digits) for _ in range(12))
        else:
            self.password = password

        # create
        self.ctr = node.spawn_container('postgres:alpine' if image is None else image,
                                        volumes=[(volume, '/var/lib/postgresql/data')],
                                        stdout_callback=log_callback)

        # async initialise
        self.async = Thread(target=self.wait_truly_up, name="Waiting for Postgres: " + self.ctr.uuid.decode())
        self.async.start()

    def wait_truly_up(self):
        try:
            # wait for TCP to come up
            self.ctr.wait_tcp(5432)

            # wait for the db to come up
            while True:
                rtn = self.ctr.run_process('psql -Upostgres -h%s -c "SELECT;"' % self.ctr.private_ip(), nolog=True)
                if rtn[2] == 0:
                    break
                logging.debug("Waiting for Postgresql to accept a query.")
                time.sleep(1)

            # actually set the password (passing it as part of the env doesn't work)
            self.ctr.run_process('psql -Upostgres -h%s -c "ALTER ROLE postgres WITH SUPERUSER PASSWORD \'%s\';"'
                                  % (self.ctr.private_ip(), self.password), nolog=True)  # prevent password being logged
            logging.info("Started Postgresql: " + self.ctr.uuid.decode())
            self.mark_as_ready()

        except BaseException as e:
            logging.critical(str(e))

    def ensure_database(self, name) -> bool:
        """Ensures a given database exists - returns True if it created a new one"""
        self.wait_until_ready()
        return self.ctr.run_process('psql -Upostgres -h%s -c "CREATE DATABASE %s WITH OWNER = postgres '
                                    'ENCODING = \'utf8\' LC_COLLATE = \'en_US.utf8\' LC_CTYPE = \'en_US.utf8\';"'
                                    % (self.ctr.private_ip(), name))[2] == 0

    def __getattr__(self, item):
        """Fake being inherited from the container"""
        return self.ctr.__getattribute__(item)

    def __repr__(self):
        return "<Postgresql '%s' pass=%s>" % (self.ctr.uuid.decode(), self.password)



# Copyright (c) 2017 David Preece, All rights reserved.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import logging
from tfnz.location import Location
from tfnz.volume import Volume
from tfnz.components.postgresql import Postgresql
from tfnz.endpoint import WebEndpoint, Cluster


class SilverStripe:
    """Puts a PHP/SilverStripe instance on each node and load balances.

    :param location: A location (object) to connect to.
    :param volume: A volume (object) to use as a persistent store.
    :param sql_volume: A volume to connect to a Postgres server for SQL storage.
    :param fqdn: The FQDN to publish to.
    :param image: Use a non-default container image.
    :param log_callback: An optional callback for log messages -  signature (object, bytes)"""

    def __init__(self, location: Location, volume: Volume, sql_volume: Volume, fqdn: str,
                 *, image=None, log_callback=None):
        nodes = location.ranked_nodes()

        # spawn database
        self.db = Postgresql(nodes[0], sql_volume, log_callback=log_callback)

        # spawn one webserver instance
        first_server = nodes[0].spawn_container('tfnz/silverstripe' if image is None else image,
                                                volumes=[(volume, '/site')],
                                                sleep=True,
                                                stdout_callback=log_callback)

        # see if the data volume has been initialised
        try:
            first_server.fetch("/site/index.php")
        except ValueError as e:
            # need to initialise - /site is visible on all servers
            logging.info("Initialising data volume: " + volume.uuid.decode())
            first_server.run_process('cp -r /silverstripe/* /site/')

        # recreate the .env because the database ip and password will have changed
        dotenv = SilverStripe.environment_template % (fqdn, self.db.password, self.db.private_ip())
        first_server.put('/site/.env', dotenv.encode())

        # start additional webservers
        self.webservers = [first_server]
        for node in nodes[1:]:
            server = node.spawn_container('tfnz/silverstripe',
                                          volumes=[(volume, '/site')],
                                          sleep=True)
            self.webservers.append(server)

        # start the actual webserving process
        fqdn_sed = "sed -i -e 's/--fqdn--/%s/g' /etc/nginx/conf.d/nginx.conf" % fqdn
        timezone_sed = "sed -i -e 's/;date.timezone =/date.timezone = %s/g' /etc/php7/php.ini" % "UTC"  # TODO
        pool_sed = "sed -i -e 's/pm.max_children = 5/pm.max_children = 16/g' /etc/php7/php-fpm.d/www.conf"
        for w in self.webservers:
            self.db.allow_connection_from(w)
            w.run_process('rm /etc/nginx/conf.d/default.conf /site/install*')
            w.run_process(fqdn_sed)
            w.run_process(timezone_sed)
            w.run_process(pool_sed)
            w.run_process('mkdir /run/nginx')
            w.spawn_process('nginx')
            w.spawn_process('php-fpm7')

        # gather together and serve into an endpoint
        self.cluster = Cluster(containers=self.webservers)
        location.endpoint_for(fqdn).publish(self.cluster, fqdn)

        # wait until we're actually able to serve
        self.db.wait_until_ready()
        WebEndpoint.wait_http_200(fqdn)
        logging.info("SilverStripe is up.")

        # Start tailing logs
        for w in self.webservers:
            w.spawn_process('tail -f /var/log/nginx/access.log', data_callback=log_callback)
            w.spawn_process('tail -f /var/log/nginx/error.log', data_callback=log_callback)

    environment_template = """
SS_BASE_URL="http://%s"
SS_DATABASE_CLASS="PostgreSQLDatabase"
SS_DATABASE_NAME="SS_mysite"
SS_DATABASE_PASSWORD="%s"
SS_DATABASE_PORT="5432"
SS_DATABASE_SERVER="%s"
SS_DATABASE_USERNAME="postgres"
SS_DEFAULT_ADMIN_USERNAME="admin"
SS_DEFAULT_ADMIN_PASSWORD="password"
"""



# Copyright (c) 2017 David Preece, All rights reserved.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import weakref
import logging
import os.path
from . import Taggable


class Volume(Taggable):
    def __init__(self, location, uuid, tag):
        super().__init__(location.user_pk, uuid, tag=tag)
        # Do not construct directly, use Location.create_volume
        self.connection = weakref.ref(location.conn)

    def snapshot(self):
        """Mark the current state of the volume as being it's initial state."""
        self.connection().send_cmd(b'snapshot_volume', {'volume': self.uuid})
        logging.info("Set snapshot for volume: " + self.uuid.decode())

    def rollback(self):
        """Resets the volume back to the initial state."""
        self.connection().send_cmd(b'rollback_volume', {'volume': self.uuid})
        logging.info("Rolled back to snapshot: " + self.uuid.decode())

    @staticmethod
    def trees_intersect(current, proposed):
        # ensure the proposed directory is neither a subdir nor a superdir of any existing directories
        p = os.path.abspath(proposed)
        for cur in current:
            c = os.path.abspath(cur)
            if len(p) > len(c):
                if p[:len(c)] == c:
                    return p, c  # p is a subtree of c
            else:
                if c[:len(p)] == p:
                    return c, p  # c is a subtree of p
        return None

    def __repr__(self):
        return "<Volume '%s'>" % self.namespaced_display_name()
